from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="TokenRateMetric")


@_attrs_define
class TokenRateMetric:
    """Token rate metric

    Attributes:
        model (Union[Unset, str]): Model ID
        provider (Union[Unset, str]): Provider name
        provider_name (Union[Unset, str]): Provider integration name
        timestamp (Union[Unset, str]): Timestamp
        token_total (Union[Unset, float]): Total tokens
        trend (Union[Unset, float]): Trend
    """

    model: Union[Unset, str] = UNSET
    provider: Union[Unset, str] = UNSET
    provider_name: Union[Unset, str] = UNSET
    timestamp: Union[Unset, str] = UNSET
    token_total: Union[Unset, float] = UNSET
    trend: Union[Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        model = self.model

        provider = self.provider

        provider_name = self.provider_name

        timestamp = self.timestamp

        token_total = self.token_total

        trend = self.trend

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if model is not UNSET:
            field_dict["model"] = model
        if provider is not UNSET:
            field_dict["provider"] = provider
        if provider_name is not UNSET:
            field_dict["providerName"] = provider_name
        if timestamp is not UNSET:
            field_dict["timestamp"] = timestamp
        if token_total is not UNSET:
            field_dict["tokenTotal"] = token_total
        if trend is not UNSET:
            field_dict["trend"] = trend

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        model = d.pop("model", UNSET)

        provider = d.pop("provider", UNSET)

        provider_name = d.pop("providerName", UNSET)

        timestamp = d.pop("timestamp", UNSET)

        token_total = d.pop("tokenTotal", UNSET)

        trend = d.pop("trend", UNSET)

        token_rate_metric = cls(
            model=model,
            provider=provider,
            provider_name=provider_name,
            timestamp=timestamp,
            token_total=token_total,
            trend=trend,
        )

        token_rate_metric.additional_properties = d
        return token_rate_metric

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import asyncio
import logging
from datetime import datetime, timedelta, timezone

import aiohttp

from blaxel.client.models import (Metadata, Port, Preview, PreviewSpec,
                                  Runtime, Sandbox, SandboxSpec)
from blaxel.common.settings import settings
from blaxel.sandbox.sandbox import SandboxInstance

logger = logging.getLogger(__name__)

async def create_sandbox(sandbox_name: str):
    # Create sandbox
    image = "blaxel/prod-base:latest"
    logger.info(f"Creating sandbox {sandbox_name} with image {image}")
    sandbox = await SandboxInstance.create(Sandbox(
        metadata=Metadata(name=sandbox_name),
        spec=SandboxSpec(
            runtime=Runtime(
                image=image,
                memory=2048,
                cpu=2,
                ports=[
                    Port(name="sandbox-api", target=8080, protocol="HTTP")
                ]
            )
        )
    ))
    logger.info("Waiting for sandbox to be deployed")
    await sandbox.wait(max_wait=120000, interval=1000)
    logger.info("Sandbox deployed")
    return sandbox

async def test_public_preview(sandbox: SandboxInstance):
    try:
        # Create a public preview
        await sandbox.previews.create(Preview(
            metadata=Metadata(name="preview-test-public"),
            spec=PreviewSpec(
                port=443,
                prefix_url="small-prefix",
                public=True
            )
        ))

        # List previews
        previews = await sandbox.previews.list()
        assert len(previews) >= 1, "No previews found"

        # Get the preview
        retrieved_preview = await sandbox.previews.get("preview-test-public")
        assert retrieved_preview.name == "preview-test-public", "Preview name is not correct"

        # Check the URL
        url = retrieved_preview.spec.url if retrieved_preview.spec else None
        assert url is not None, "Preview URL is not correct"
        workspace = settings.workspace
        expected_url = f"https://small-prefix-{workspace}.preview.bl.run"
        assert url == expected_url, f"Preview URL is not correct => {url}"

        # Test the preview endpoint
        async with aiohttp.ClientSession() as session:
            async with session.get(f"{url}/health") as response:
                assert response.status == 200, f"Preview is not working => {response.status}:{await response.text()}"

        logger.info("Public preview is healthy :)")
    except Exception as e:
        logger.error("ERROR IN PUBLIC PREVIEW TEST => ", exc_info=e)
        raise
    finally:
        await sandbox.previews.delete("preview-test-public")

async def test_private_preview(sandbox: SandboxInstance):
    try:
        # Create a private preview
        preview = await sandbox.previews.create(Preview(
            metadata=Metadata(name="preview-test-private"),
            spec=PreviewSpec(
                port=443,
                public=False
            )
        ))

        # Get the preview URL
        url = preview.spec.url if preview.spec else None
        assert url is not None, "Preview URL is not correct"

        # Create a token
        token = await preview.tokens.create(datetime.now(timezone.utc) + timedelta(minutes=10))
        logger.info(f"Token created => {token.value}")

        # List tokens
        tokens = await preview.tokens.list()
        assert len(tokens) >= 1, "No tokens found"
        assert any(t.value == token.value for t in tokens), "Token not found in list"

        # Test the preview endpoint without token
        async with aiohttp.ClientSession() as session:
            async with session.get(f"{url}/health") as response:
                assert response.status == 401, f"Preview is not protected by token, response => {response.status}"

        # Test the preview endpoint with token
        async with aiohttp.ClientSession() as session:
            async with session.get(f"{url}/health?bl_preview_token={token.value}") as response:
                assert response.status == 200, f"Preview is not working with token, response => {response.status}"

        logger.info("Private preview is healthy with token :)")

        # Delete the token
        await preview.tokens.delete(token.value)
    except Exception as e:
        logger.error("ERROR IN PRIVATE PREVIEW TEST => ", exc_info=e)
        raise
    finally:
        await sandbox.previews.delete("preview-test-private")

async def main():
    sandbox_name = "sandbox-preview-test"
    sandbox = None
    try:
        sandbox = await create_sandbox(sandbox_name)
        await test_public_preview(sandbox)
        await test_private_preview(sandbox)
    finally:
        if sandbox:
            logger.info("Deleting sandbox")
            await SandboxInstance.delete(sandbox_name)

if __name__ == "__main__":
    asyncio.run(main())


from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.store_configuration_option import StoreConfigurationOption


T = TypeVar("T", bound="StoreConfiguration")


@_attrs_define
class StoreConfiguration:
    """Store configuration for resources (eg: agent, function, etc)

    Attributes:
        available_models (Union[Unset, list[str]]): Available models for the configuration
        description (Union[Unset, str]): Store configuration description
        display_name (Union[Unset, str]): Store configuration display name
        if_ (Union[Unset, str]): Conditional rendering for the configuration, example: provider === 'openai'
        name (Union[Unset, str]): Store configuration name
        options (Union[Unset, list['StoreConfigurationOption']]):
        required (Union[Unset, bool]): Store configuration required
        secret (Union[Unset, bool]): Store configuration secret
        type_ (Union[Unset, str]): Store configuration type
    """

    available_models: Union[Unset, list[str]] = UNSET
    description: Union[Unset, str] = UNSET
    display_name: Union[Unset, str] = UNSET
    if_: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    options: Union[Unset, list["StoreConfigurationOption"]] = UNSET
    required: Union[Unset, bool] = UNSET
    secret: Union[Unset, bool] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        available_models: Union[Unset, list[str]] = UNSET
        if not isinstance(self.available_models, Unset):
            available_models = self.available_models

        description = self.description

        display_name = self.display_name

        if_ = self.if_

        name = self.name

        options: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.options, Unset):
            options = []
            for options_item_data in self.options:
                if type(options_item_data) == dict:
                    options_item = options_item_data
                else:
                    options_item = options_item_data.to_dict()
                options.append(options_item)

        required = self.required

        secret = self.secret

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if available_models is not UNSET:
            field_dict["availableModels"] = available_models
        if description is not UNSET:
            field_dict["description"] = description
        if display_name is not UNSET:
            field_dict["displayName"] = display_name
        if if_ is not UNSET:
            field_dict["if"] = if_
        if name is not UNSET:
            field_dict["name"] = name
        if options is not UNSET:
            field_dict["options"] = options
        if required is not UNSET:
            field_dict["required"] = required
        if secret is not UNSET:
            field_dict["secret"] = secret
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.store_configuration_option import StoreConfigurationOption

        if not src_dict:
            return None
        d = src_dict.copy()
        available_models = cast(list[str], d.pop("availableModels", UNSET))

        description = d.pop("description", UNSET)

        display_name = d.pop("displayName", UNSET)

        if_ = d.pop("if", UNSET)

        name = d.pop("name", UNSET)

        options = []
        _options = d.pop("options", UNSET)
        for options_item_data in _options or []:
            options_item = StoreConfigurationOption.from_dict(options_item_data)

            options.append(options_item)

        required = d.pop("required", UNSET)

        secret = d.pop("secret", UNSET)

        type_ = d.pop("type", UNSET)

        store_configuration = cls(
            available_models=available_models,
            description=description,
            display_name=display_name,
            if_=if_,
            name=name,
            options=options,
            required=required,
            secret=secret,
            type_=type_,
        )

        store_configuration.additional_properties = d
        return store_configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
import os
import time

from blaxel.client import client
from blaxel.client.api.functions import create_function, delete_function, get_function
from blaxel.client.api.integrations import (
    create_integration_connection,
    delete_integration_connection,
)
from blaxel.client.models import (
    Function,
    FunctionSpec,
    IntegrationConnection,
    IntegrationConnectionSpec,
    Metadata,
    Runtime,
)
from blaxel.tools import bl_tools

logger = logging.getLogger(__name__)

async def wait_mcp_ready(name, max_wait: int = 60000, interval: int = 1000):
    start_time = time.time() * 1000
    mcp: Function = await get_function.asyncio(
        name,
        client=client,
    )
    while mcp.status != "DEPLOYED":
        await asyncio.sleep(1)
        logger.info(f"Waiting for MCP to be deployed, status: {mcp.status}")
        mcp = await get_function.asyncio(
            name,
            client=client,
        )
        if (time.time() * 1000) - start_time > max_wait:
            raise Exception("MCP did not deploy in time")

async def main():
    try:
        team_id = os.getenv("SLACK_TEAM_ID")
        bot_token = os.getenv("SLACK_BOT_TOKEN")
        mcp_name = "my-slack-mcp"
        integration_connection_name = "my-slack-integration"

        if not team_id or not bot_token:
            raise Exception("SLACK_TEAM_ID and SLACK_BOT_TOKEN must be set")

        # Create integration connection
        integration_connection: IntegrationConnection = await create_integration_connection.asyncio(
            client=client,
            body=IntegrationConnection(
                metadata=Metadata(
                    name=integration_connection_name
                ),
                spec=IntegrationConnectionSpec(
                    integration="slack",
                    config={
                        "teamId": team_id
                    },
                    secret={
                        "botToken": bot_token
                    }
                )
            )
        )

        # Create MCP
        await create_function.asyncio(
            client=client,
            body=Function(
                metadata=Metadata(
                    name=mcp_name
                ),
                spec=FunctionSpec(
                    integration_connections=[integration_connection.metadata.name],
                    runtime=Runtime(
                        type_="mcp"
                    )
                )
            )
        )
        await wait_mcp_ready(mcp_name)

        ### DO SOME STUFF WITH MCP
        tools = await bl_tools([mcp_name]).initialize()
        print(await tools.get_tools()[0].coroutine())
    except Exception as e:
        logger.error(f"There was an error while creating/using MCP, {e}")
    finally:
        ### CLEANUP
        await delete_function.asyncio(
            mcp_name,
            client=client,
        )
        await delete_integration_connection.asyncio(
            integration_connection_name,
            client=client,
        )

if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



import nest_asyncio

# This is because crewai is not asyncio compatible for now: https://github.com/crewAIInc/crewAI/issues/195
nest_asyncio.apply()

import time
from contextlib import asynccontextmanager
from logging import getLogger

import uvicorn
from asgi_correlation_id import CorrelationIdMiddleware
from crewai import Agent, Crew, Task
from fastapi import FastAPI, Request, Response
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

from blaxel.instrumentation.span import SpanManager
from blaxel.models import bl_model
from blaxel.tools import bl_tools

logger = getLogger(__name__)

@asynccontextmanager
async def lifespan(app: FastAPI):
    logger.info("Server running on port 1338")
    yield
    logger.info("Server shutting down")

app = FastAPI(lifespan=lifespan)
app.add_middleware(CorrelationIdMiddleware)

@app.middleware("http")
async def log_requests(request: Request, call_next):
    start_time = time.time()

    response: Response = await call_next(request)

    process_time = (time.time() - start_time) * 1000
    formatted_process_time = f'{process_time:.2f}'
    rid_header = response.headers.get("X-Request-Id")
    request_id = rid_header or response.headers.get("X-Blaxel-Request-Id")
    logger.info(f"{request.method} {request.url.path} {response.status_code} {formatted_process_time}ms rid={request_id}")

    return response

@app.post("/")
async def handle_request(request: Request):
    tools = await bl_tools(["blaxel-search"]).to_crewai()
    model = await bl_model("gpt-4o-mini").to_crewai()
    agent = Agent(
        role="Weather Researcher",
        goal="Find the weather in a specific city",
        backstory="You are an experienced weather researcher with attention to detail",
        llm=model,
        tools=tools,
        verbose=True,
    )
    with SpanManager("blaxel-crewai").create_active_span("agent-request", {}) as span:
        body = await request.json()
        crew = Crew(
            agents=[agent],
            tasks=[
                Task(
                    description=body.get("description", "Find the weather in a specific city"),
                    expected_output=body.get("expected_output", "Weather in San francisco."),
                    agent=agent
                )
            ],
            verbose=True,
        )
        result = crew.kickoff()
        span.set_attribute("agent.result", result.raw)
        return Response(
            result.raw,
            media_type="text/plain"
        )

FastAPIInstrumentor.instrument_app(app)

if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=1338, log_level="critical", loop="asyncio")



"""Cache module for storing and retrieving resources."""

import os
from typing import Any, Dict, Optional

import yaml

# Global cache to store resources
_cache: Dict[str, Any] = {}


def _load_cache() -> None:
    """
    Load cache data from .cache.yaml file if it exists.
    """
    try:
        if os.path.exists('.cache.yaml'):
            with open('.cache.yaml', encoding='utf-8') as file:
                cache_string = file.read()
                cache_data = yaml.safe_load_all(cache_string)

                for doc in cache_data:
                    if doc and isinstance(doc, dict) and 'kind' in doc and 'metadata' in doc and 'name' in doc['metadata']:
                        cache_key = f"{doc['kind']}/{doc['metadata']['name']}"
                        _cache[cache_key] = doc
    except Exception:
        # Silently fail if cache file doesn't exist or is invalid
        pass

# Initialize cache on module import
_load_cache()


async def find_from_cache(resource: str, name: str) -> Optional[Any]:
    """
    Find a resource from the cache by resource type and name.

    Args:
        resource: The resource type
        name: The resource name

    Returns:
        The cached resource or None if not found
    """
    cache_key = f"{resource}/{name}"
    return _cache.get(cache_key)



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.function_schema import FunctionSchema


T = TypeVar("T", bound="FunctionKit")


@_attrs_define
class FunctionKit:
    """Function kit

    Attributes:
        description (Union[Unset, str]): Description of the function kit, very important for the agent to work with your
            kit
        name (Union[Unset, str]): The kit name, very important for the agent to work with your kit
        schema (Union[Unset, FunctionSchema]): Function schema
    """

    description: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    schema: Union[Unset, "FunctionSchema"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        name = self.name

        schema: Union[Unset, dict[str, Any]] = UNSET
        if self.schema and not isinstance(self.schema, Unset) and not isinstance(self.schema, dict):
            schema = self.schema.to_dict()
        elif self.schema and isinstance(self.schema, dict):
            schema = self.schema

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if description is not UNSET:
            field_dict["description"] = description
        if name is not UNSET:
            field_dict["name"] = name
        if schema is not UNSET:
            field_dict["schema"] = schema

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.function_schema import FunctionSchema

        if not src_dict:
            return None
        d = src_dict.copy()
        description = d.pop("description", UNSET)

        name = d.pop("name", UNSET)

        _schema = d.pop("schema", UNSET)
        schema: Union[Unset, FunctionSchema]
        if isinstance(_schema, Unset):
            schema = UNSET
        else:
            schema = FunctionSchema.from_dict(_schema)

        function_kit = cls(
            description=description,
            name=name,
            schema=schema,
        )

        function_kit.additional_properties = d
        return function_kit

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This module provides a custom colored formatter for logging and an initialization function
to set up logging configurations for Blaxel applications.
"""

import logging


class ColoredFormatter(logging.Formatter):
    """
    A custom logging formatter that adds ANSI color codes to log levels for enhanced readability.

    Attributes:
        COLORS (dict): A mapping of log level names to their corresponding ANSI color codes.
    """
    COLORS = {
        "DEBUG": "\033[1;36m",  # Cyan
        "INFO": "\033[1;32m",  # Green
        "WARNING": "\033[1;33m",  # Yellow
        "ERROR": "\033[1;31m",  # Red
        "CRITICAL": "\033[1;41m",  # Red background
    }

    def format(self, record):
        """
        Formats the log record by adding color codes based on the log level.

        Parameters:
            record (LogRecord): The log record to format.

        Returns:
            str: The formatted log message with appropriate color codes.
        """
        n_spaces = len("CRITICAL") - len(record.levelname)
        tab = " " * n_spaces
        color = self.COLORS.get(record.levelname, "\033[0m")
        record.levelname = f"{color}{record.levelname}\033[0m:{tab}"
        return super().format(record)

def init_logger(log_level: str):
    """
    Initializes the logging configuration for Blaxel.

    This function clears existing handlers for specific loggers, sets up a colored formatter,
    and configures the root logger with the specified log level.

    Parameters:
        log_level (str): The logging level to set (e.g., "DEBUG", "INFO").
    """
    # Disable urllib3 logging
    logging.getLogger('urllib3').setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.CRITICAL)
    handler = logging.StreamHandler()
    handler.setFormatter(ColoredFormatter("%(levelname)s %(name)s - %(message)s"))
    logging.basicConfig(level=log_level, handlers=[handler])



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ProcessRequest")


@_attrs_define
class ProcessRequest:
    """
    Attributes:
        command (str):  Example: ls -la.
        name (Union[Unset, str]):  Example: my-process.
        timeout (Union[Unset, int]):  Example: 30.
        wait_for_completion (Union[Unset, bool]):
        wait_for_ports (Union[Unset, list[int]]):  Example: [3000, 8080].
        working_dir (Union[Unset, str]):  Example: /home/user.
    """

    command: str
    name: Union[Unset, str] = UNSET
    timeout: Union[Unset, int] = UNSET
    wait_for_completion: Union[Unset, bool] = UNSET
    wait_for_ports: Union[Unset, list[int]] = UNSET
    working_dir: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        command = self.command

        name = self.name

        timeout = self.timeout

        wait_for_completion = self.wait_for_completion

        wait_for_ports: Union[Unset, list[int]] = UNSET
        if not isinstance(self.wait_for_ports, Unset):
            wait_for_ports = self.wait_for_ports

        working_dir = self.working_dir

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "command": command,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if timeout is not UNSET:
            field_dict["timeout"] = timeout
        if wait_for_completion is not UNSET:
            field_dict["waitForCompletion"] = wait_for_completion
        if wait_for_ports is not UNSET:
            field_dict["waitForPorts"] = wait_for_ports
        if working_dir is not UNSET:
            field_dict["workingDir"] = working_dir

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        command = d.pop("command")

        name = d.pop("name", UNSET)

        timeout = d.pop("timeout", UNSET)

        wait_for_completion = d.pop("waitForCompletion", UNSET)

        wait_for_ports = cast(list[int], d.pop("waitForPorts", UNSET))

        working_dir = d.pop("workingDir", UNSET)

        process_request = cls(
            command=command,
            name=name,
            timeout=timeout,
            wait_for_completion=wait_for_completion,
            wait_for_ports=wait_for_ports,
            working_dir=working_dir,
        )

        process_request.additional_properties = d
        return process_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os
from typing import Dict, Optional

import tomli
from pydantic import BaseModel


class EnvConfig(BaseModel):
    secret_env: Dict[str, str] = {}
    config_env: Dict[str, str] = {}

    def __getattr__(self, name: str) -> Optional[str]:
        if name in self.secret_env:
            return self.secret_env[name]
        if name in self.config_env:
            return self.config_env[name]
        return os.environ.get(name)

    def __getitem__(self, name: str) -> Optional[str]:
        return self.__getattr__(name)

def load_env() -> EnvConfig:
    env_config = EnvConfig()

    # Load config from blaxel.toml
    try:
        with open('blaxel.toml', 'rb') as f:
            config_infos = tomli.load(f)
            if 'env' in config_infos:
                env_config.config_env.update(config_infos['env'])
    except Exception:
        pass

    # Load secrets from .env
    try:
        with open('.env') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#'):
                    key, value = line.split('=', 1)
                    env_config.secret_env[key] = value.replace('"', '')
    except Exception:
        pass

    return env_config

env = load_env()


import inspect
from typing import Any, Optional, override

from google.adk.tools import BaseTool, ToolContext
from google.genai import types

from .types import Tool


def get_google_adk_tools(tools: list[Tool]) -> list[BaseTool]:
    class GoogleADKTool(BaseTool):
        _tool: Tool

        def __init__(self, tool: Tool):
            super().__init__(
                name=tool.name,
                description=tool.description,
            )
            self._tool = tool

        def _clean_schema(self, schema: dict) -> dict:
            if not isinstance(schema, dict):
                return schema

            # Create a copy of the schema
            cleaned_schema = schema.copy()

            # Remove $schema and additionalProperties at current level
            if "$schema" in cleaned_schema:
                del cleaned_schema["$schema"]
            if "additionalProperties" in cleaned_schema:
                del cleaned_schema["additionalProperties"]

            # Recursively clean properties if they exist
            if "properties" in cleaned_schema:
                cleaned_schema["properties"] = {
                    k: self._clean_schema(v) for k, v in cleaned_schema["properties"].items()
                }

            return cleaned_schema

        @override
        def _get_declaration(self) -> Optional[types.FunctionDeclaration]:
            # Clean the schema recursively
            schema = self._clean_schema(self._tool.input_schema)

            function_decl = types.FunctionDeclaration.model_validate(
                types.FunctionDeclaration(
                    name=self._tool.name,
                    description=self._tool.description,
                    parameters=schema,
                )
            )

            return function_decl

        @override
        async def run_async(
            self, *, args: dict[str, Any], tool_context: ToolContext
        ) -> Any:
            args_to_call = args.copy()
            signature = inspect.signature(self._tool.coroutine)
            if 'tool_context' in signature.parameters:
                args_to_call['tool_context'] = tool_context
            return await self._tool.coroutine(**args_to_call) or {}
    return [GoogleADKTool(tool) for tool in tools]


from agents import AsyncOpenAI, OpenAIChatCompletionsModel

from ..common.settings import settings


async def get_openai_model(url: str, type: str, model: str, **kwargs):
    if type != "openai":
        raise ValueError(f"Invalid model type: {type}")
    external_client = AsyncOpenAI(
        base_url=f"{url}/v1",
        api_key=settings.auth.token,
        default_headers=settings.headers,
    )

    return OpenAIChatCompletionsModel(
        model=model,
        openai_client=external_client,
        **kwargs
    )



from ..cache import find_from_cache
from ..client import client
from ..client.api.models import get_model
from ..client.models import Model
from ..common.settings import settings


class BLModel:
    models = {}

    def __init__(self, model_name, **kwargs):
        self.model_name = model_name
        self.kwargs = kwargs

    async def to_langchain(self):
        from .langchain import get_langchain_model
        url, type, model = await self._get_parameters()
        model = await get_langchain_model(url, type, model, **self.kwargs)
        BLModel.models[f"langchain_{self.model_name}"] = model
        return model

    async def to_llamaindex(self):
        from .llamaindex import get_llamaindex_model
        url, type, model = await self._get_parameters()
        model = await get_llamaindex_model(url, type, model, **self.kwargs)
        BLModel.models[f"llamaindex_{self.model_name}"] = model
        return model

    async def to_crewai(self):
        from .crewai import get_crewai_model

        url, type, model = await self._get_parameters()
        model = await get_crewai_model(url, type, model, **self.kwargs)
        BLModel.models[f"crewai_{self.model_name}"] = model
        return model

    async def to_openai(self):
        from .openai import get_openai_model
        url, type, model = await self._get_parameters()
        model = await get_openai_model(url, type, model, **self.kwargs)
        BLModel.models[f"openai_{self.model_name}"] = model
        return model

    async def to_pydantic(self):
        from .pydantic import get_pydantic_model
        url, type, model = await self._get_parameters()
        model = await get_pydantic_model(url, type, model, **self.kwargs)
        BLModel.models[f"pydantic_{self.model_name}"] = model
        return model

    async def to_google_adk(self):
        from .googleadk import get_google_adk_model
        url, type, model = await self._get_parameters()
        model = await get_google_adk_model(url, type, model, **self.kwargs)
        BLModel.models[f"googleadk_{self.model_name}"] = model
        return model

    async def to_livekit(self):
        # This has to be here because livekit plugins must be registered on the main thread
        from .livekit import get_livekit_model

        url, type, model = await self._get_parameters()
        model = await get_livekit_model(url, type, model, **self.kwargs)
        BLModel.models[f"livekit_{self.model_name}"] = model
        return model

    async def _get_parameters(self) -> tuple[str, str, str]:
        if self.model_name in self.models:
            # We get the headers in case we need to refresh the token
            settings.auth.get_headers()
            model = self.models[self.model_name]
            return model["url"], model["type"], model["model"]
        url = f"{settings.run_url}/{settings.auth.workspace_name}/models/{self.model_name}"
        model_data = await self._get_model_metadata()
        if not model_data:
            raise Exception(f"Model {self.model_name} not found")
        runtime = (model_data.spec and model_data.spec.runtime)
        if not runtime:
            raise Exception(f"Model {self.model_name} has no runtime")

        type = runtime.type_ or 'openai'
        model = runtime.model
        self.models[self.model_name] = {
            "url": url,
            "type": type,
            "model": model
        }
        return url, type, model

    async def _get_model_metadata(self) -> Model | None:
        cache_data = await find_from_cache('Model', self.model_name)
        if cache_data:
            return Model.from_dict(cache_data)

        try:
            return await get_model.asyncio(client=client, model_name=self.model_name)
        except Exception:
            return None

def bl_model(model_name, **kwargs):
    return BLModel(model_name, **kwargs)

__all__ = ["bl_model"]



import asyncio
import time
from collections import Counter

from blaxel.client import client
from blaxel.common.settings import settings


async def request_agent(name):
    response = await client.get_async_httpx_client().get(
        f"{settings.run_url}/{settings.workspace}/agents/{name}"
    )
    if response.status_code != 200:
        raise Exception(f"Agent {name} returned status code {response.status_code} with body {response.text}")
    return response.status_code

async def main():
    number_of_calls = 500
    batch_size = 100
    agent_name = "agent-telemetry-ts"

    total_start_time = time.time()
    batch_times = []

    for batch_start in range(0, number_of_calls, batch_size):
        batch_end = min(batch_start + batch_size, number_of_calls)
        batch_tasks = [request_agent(agent_name) for _ in range(batch_start, batch_end)]
        print(f"Processing batch {batch_start//batch_size + 1} ({batch_start}-{batch_end-1})")

        batch_start_time = time.time()
        status_codes = await asyncio.gather(*batch_tasks)
        batch_time = time.time() - batch_start_time
        batch_times.append(batch_time)

        status_counts = Counter(status_codes)
        print(f"Completed batch {batch_start//batch_size + 1}")
        print(f"Status codes: {dict(status_counts)}")
        print(f"Batch time: {batch_time:.2f}s")
        print(f"Mean time per request in batch: {batch_time/batch_size:.3f}s")
        print("-" * 50)

    total_time = time.time() - total_start_time
    mean_time_per_request = total_time / number_of_calls

    print("\nSummary:")
    print(f"Total execution time: {total_time:.2f}s")
    print(f"Mean time per request: {mean_time_per_request:.3f}s")
    print(f"Fastest batch: {min(batch_times):.2f}s")
    print(f"Slowest batch: {max(batch_times):.2f}s")
    print(f"Average batch time: {sum(batch_times)/len(batch_times):.2f}s")

if __name__ == "__main__":
    asyncio.run(main())



import logging

from anthropic import AsyncAnthropic
from cohere import AsyncClientV2
from mistralai.sdk import Mistral
from pydantic_ai.models import Model
from pydantic_ai.models.anthropic import AnthropicModel
from pydantic_ai.models.cohere import CohereModel
from pydantic_ai.models.gemini import GeminiModel
from pydantic_ai.models.mistral import MistralModel
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.anthropic import AnthropicProvider
from pydantic_ai.providers.cohere import CohereProvider
from pydantic_ai.providers.mistral import MistralProvider
from pydantic_ai.providers.openai import OpenAIProvider

from blaxel.common.settings import settings

from ..client import client
from .custom.pydantic.gemini import GoogleGLAProvider

logger = logging.getLogger(__name__)
async def get_pydantic_model(url: str,type: str, model: str, **kwargs) -> Model:
    if type == 'mistral':
        return MistralModel(
            model_name=model,
            provider=MistralProvider(
                mistral_client=Mistral(
                    api_key=settings.auth.token,
                    server_url=url,
                ),
                **kwargs
            ),
        )
    elif type == 'cohere':
        return CohereModel(
            model_name=model,
            provider=CohereProvider(
                cohere_client=AsyncClientV2(
                    api_key=settings.auth.token,
                    base_url=url,
                ),
            ),
        )
    elif type == 'xai':
        return OpenAIModel(
            model_name=model,
            provider=OpenAIProvider(
                base_url=f"{url}/v1",
                api_key=settings.auth.token,
                **kwargs
            ),
        )
    elif type == 'deepseek':
        return OpenAIModel(
            model_name=model,
            provider=OpenAIProvider(
                base_url=f"{url}/v1",
                api_key=settings.auth.token,
                **kwargs
            ),
        )
    elif type == 'cerebras':
        return OpenAIModel(
            model_name=model,
            provider=OpenAIProvider(
                base_url=f"{url}/v1",
                api_key=settings.auth.token,
                **kwargs
            ),
        )
    elif type == 'anthropic':
        return AnthropicModel(
            model_name=model,
            provider=AnthropicProvider(
                anthropic_client=AsyncAnthropic(
                    api_key=settings.auth.token,
                    base_url=url,
                    default_headers=settings.auth.get_headers(),
                ),
                **kwargs
            )
        )
    elif type == 'gemini':
        return GeminiModel(
            model_name=model,
            provider=GoogleGLAProvider(
                api_key=settings.auth.token,
                http_client=client.with_base_url(f"{url}/v1beta/models").get_async_httpx_client()
            )
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by Pydantic, defaulting to OpenAI")
        return OpenAIModel(
            model_name=model,
            provider=OpenAIProvider(
                base_url=f"{url}/v1",
                api_key=settings.auth.token,
                **kwargs
            ),
        )


from blaxel.sandbox.client.models.process_request import ProcessRequest

from .base import SandboxHandleBase
from .client.api.process.delete_process_identifier import (
    asyncio_detailed as delete_process_by_identifier,
)
from .client.api.process.delete_process_identifier_kill import (
    asyncio_detailed as delete_process_by_identifier_kill,
)
from .client.api.process.get_process import asyncio_detailed as get_process
from .client.api.process.get_process_identifier import asyncio_detailed as get_process_by_identifier
from .client.api.process.get_process_identifier_logs import (
    asyncio_detailed as get_process_by_identifier_logs,
)
from .client.api.process.post_process import asyncio_detailed as post_process
from .client.models import (
    GetProcessIdentifierLogsResponse200,
    ProcessKillRequest,
    ProcessResponse,
    SuccessResponse,
)


class SandboxProcess(SandboxHandleBase):
    async def exec(self, process: ProcessRequest) -> ProcessResponse:
        response = await post_process(client=self.client, body=process)
        self.handle_response(response)
        return response.parsed

    async def get(self, identifier: str) -> ProcessResponse:
        response = await get_process_by_identifier(identifier=identifier, client=self.client)
        self.handle_response(response)
        return response.parsed

    async def list(self) -> list[ProcessResponse]:
        response = await get_process(client=self.client)
        self.handle_response(response)
        return response.parsed

    async def stop(self, identifier: str) -> SuccessResponse:
        response = await delete_process_by_identifier(identifier=identifier, client=self.client)
        self.handle_response(response)
        return response.parsed

    async def kill(self, identifier: str, signal: str = "SIGKILL") -> SuccessResponse:
        kill_request = ProcessKillRequest(signal=signal)
        response = await delete_process_by_identifier_kill(identifier=identifier, client=self.client, body=kill_request)
        self.handle_response(response)
        return response.parsed

    async def logs(self, identifier: str, type_: str = "stdout") -> str:
        response = await get_process_by_identifier_logs(identifier=identifier, client=self.client)
        self.handle_response(response)
        data: GetProcessIdentifierLogsResponse200 = response.parsed
        if type_ in data.additional_properties:
            return data.additional_properties[type_]
        raise Exception("Unsupported log type")



import asyncio
import logging
import os

from blaxel.client.models import Metadata, Port, Runtime, Sandbox, SandboxSpec
from blaxel.sandbox.base import ResponseError
from blaxel.sandbox.client.models import ProcessRequest
from blaxel.sandbox.sandbox import SandboxInstance

logger = logging.getLogger(__name__)


async def create_sandbox(sandbox_name: str):
    # Create sandbox
    image = "blaxel/prod-base:latest"
    logger.info(f"Creating sandbox {sandbox_name} with image {image}")
    sandbox = await SandboxInstance.create(Sandbox(
        metadata=Metadata(name=sandbox_name),
        spec=SandboxSpec(
            runtime=Runtime(
                image=image,
                memory=2048,
                cpu=2,
                ports=[
                    Port(name="sandbox-api", target=8080, protocol="HTTP")
                ]
            )
        )
    ))
    logger.info("Waiting for sandbox to be deployed")
    await sandbox.wait(max_wait=120000, interval=1000)
    logger.info("Sandbox deployed")
    return sandbox

async def main():
    user = os.environ.get("USER")
    sandbox_name = "sandbox-test-3"
    sandbox = None
    try:
        sandbox = await create_sandbox(sandbox_name)
        sandbox = await SandboxInstance.get(sandbox_name)

        # Filesystem tests
        await sandbox.fs.write(f"/Users/{user}/Downloads/test", "Hello world")
        content = await sandbox.fs.read(f"/Users/{user}/Downloads/test")
        assert content == "Hello world", "File content is not correct"
        dir = await sandbox.fs.ls(f"/Users/{user}/Downloads")
        assert dir.files and len(dir.files) >= 1, "Directory is empty"
        assert any(f.path == f"/Users/{user}/Downloads/test" for f in dir.files), "File not found in directory"
        await sandbox.fs.mkdir(f"/Users/{user}/Downloads/test2")
        after_mkdir = await sandbox.fs.ls(f"/Users/{user}/Downloads/test2")
        assert not after_mkdir.files or len(after_mkdir.files) == 0, "Directory is not empty after mkdir"
        await sandbox.fs.cp(f"/Users/{user}/Downloads/test", f"/Users/{user}/Downloads/test2/test")
        after_cp_ls = await sandbox.fs.ls(f"/Users/{user}/Downloads/test2")
        assert after_cp_ls.files and any(f.path == f"/Users/{user}/Downloads/test2/test" for f in after_cp_ls.files), "File not found in directory after cp"
        await sandbox.fs.rm(f"/Users/{user}/Downloads/test")
        try:
            await sandbox.fs.rm(f"/Users/{user}/Downloads/test2")
        except ResponseError as e:
            logger.info(f"That is expected => {e.error}")
        await sandbox.fs.rm(f"/Users/{user}/Downloads/test2", True)

        # Process tests
        process = await sandbox.process.exec(ProcessRequest(name="test", command="echo 'Hello world'"))
        assert getattr(process, "status", None) != "completed", "Process did complete without waiting"
        await asyncio.sleep(0.01)
        completed_process = await sandbox.process.get("test")
        assert getattr(completed_process, "status", None) == "completed", "Process did not complete"
        logs = await sandbox.process.logs("test")
        assert logs == 'Hello world\n', "Logs are not correct"
        try:
            await sandbox.process.kill("test")
        except ResponseError as e:
            logger.info(f"That is expected => {e.error}")
    finally:
        logger.info("Deleting sandbox")
        await SandboxInstance.delete(sandbox_name)

if __name__ == "__main__":
    asyncio.run(main())



from logging import getLogger

from livekit.plugins import groq, openai

from ..common.settings import settings

logger = getLogger(__name__)

async def get_livekit_model(url: str, type: str, model: str, **kwargs):
    if type == 'xai':
        return groq.LLM(
            model=model,
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )
    else:
        if type != 'openai':
            logger.warning(f"Livekit not compatible with: {type}, defaulting to openai.LLM")
        return openai.LLM(
            model=model,
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )



from logging import getLogger

from crewai import LLM

from ..common.settings import settings

logger = getLogger(__name__)

async def get_crewai_model(url: str, type: str, model: str, **kwargs):
    if type == 'mistral':
        return LLM(
            model=f"mistral/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )
    elif type == 'xai':
        return LLM(
            model=f"groq/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )
    elif type == 'deepseek':
        return LLM(
            model=f"openai/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )
    elif type == 'anthropic':
        return LLM(
            model=f"anthropic/{model}",
            api_key=settings.auth.token,
            base_url=url,
            extra_headers=settings.auth.get_headers(),
            **kwargs
        )
    elif type == 'gemini':
        return LLM(
            model=f"gemini/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1beta/models/{model}",
            **kwargs
        )
    elif type == 'cerebras':
        return LLM(
            model=f"cerebras/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )
    else:
        if type != "openai":
            logger.warning(f"Model {model} is not supported by CrewAI, defaulting to OpenAI")
        return LLM(
            model=f"openai/{model}",
            api_key=settings.auth.token,
            base_url=f"{url}/v1",
            **kwargs
        )



import asyncio
import json
import logging

from blaxel.sandbox.sandbox import SandboxInstance

logger = logging.getLogger(__name__)


async def main():
    sandbox_name = "spinup-test5"
    sandbox = await SandboxInstance.get(sandbox_name)
    result = await sandbox.fs.ls("/root")
    print(json.dumps(result.to_dict(), indent=4))
    # Filesystem tests

if __name__ == "__main__":
    asyncio.run(main())



import json
from typing import Any

from agents import FunctionTool, RunContextWrapper

from .types import Tool


def get_openai_tool(tool: Tool) -> FunctionTool:
    async def openai_coroutine(
      _: RunContextWrapper,
      arguments: dict[str, Any],
    ) -> Any:
        result = await tool.coroutine(**json.loads(arguments))
        return result

    return FunctionTool(
        name=tool.name,
        description=tool.description,
        params_json_schema=tool.input_schema,
        on_invoke_tool=openai_coroutine,
    )

def get_openai_tools(tools: list[Tool]) -> list[FunctionTool]:
    return [get_openai_tool(tool) for tool in tools]
